/*
Copyright Buddies of Budgie

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
	Yes that Apache-2.0 block is larger than the CSS itself.
*/

.desktop-item-label.larger-text {
  /* Icon Size is large or massive */
  font-size: 14px;
}

.desktop-item-label.is-disabled {
  /* Is the item currently being copied */
  color: rgba(232, 232, 232, 0.7);
}

.desktop-item:not(.selected) .desktop-item-image {
  opacity: 0.9;
}

.desktop-item:not(.selected) .desktop-item-label {
  color: rgba(232, 232, 232, 0.9);
  font-weight: 600;
  text-shadow: 0 0 2px rgba(0, 0, 0, 1), 0 0 2px rgba(0, 0, 0, 1);
}

.desktop-item.selected .desktop-item-label {
  color: rgba(240, 240, 240, 1);
  font-weight: 600;
  text-shadow: 0 0 2px rgba(0, 0, 0, 1), 0 0 2px rgba(0, 0, 0, 1);
}

.debug {
  background: white;
}
